@@ -1,15 +1,19 @@
-#!/usr/bin/env python3
+#!./runmodule.sh
 
 '''
 An editor for CAN bus database files like dbc or sym, based on cantools and urwid
 '''
 
-__version__ = '0.1.0'
+__version__ = '0.3.0'
 
 
 import os
 import sys
 import argparse
+import tempfile
+import logging
+import abc
+from gettext import gettext as _
 
 import urwid
 import cantools
@@ -24,13 +28,21 @@ ATTR_ERROR = 'error'
 
 CANCEL = 'cancel'
 
+urwid.command_map['j'] = urwid.CURSOR_DOWN
+urwid.command_map['k'] = urwid.CURSOR_UP
+urwid.command_map['h'] = urwid.CURSOR_LEFT
+urwid.command_map['l'] = urwid.CURSOR_RIGHT
+
 
 # ------- base classes -------
 
+# I am not inheriting from abc.ABC to avoid TypeError: metaclass conflict: the metaclass of a derived class must be a (non-strict) subclass of the metaclasses of all its bases
 class MyWindow(urwid.Frame):
 
 	def __init__(self, body=None, **kw):
 		self.prev_window = kw.pop('prev_window', None)
+		self.errors: 'list[str]' = []
+		self.help_line = None
 		super().__init__(body, **kw)
 
 	def set_body(self, body):
@@ -42,7 +54,8 @@ class MyWindow(urwid.Frame):
 	def set_help_line(self, shortcuts):
 		'''shortcuts: iterable of (key, command) tuples'''
 		ln = ", ".join(f"{key}: {cmd}" for key, cmd in shortcuts)
-		self.footer = urwid.Text((ATTR_HELP_LINE, ln))
+		self.help_line = urwid.Text((ATTR_HELP_LINE, ln))
+		self.footer = self.help_line
 
 	def cmd_back(self):
 		return self.prev_window
@@ -63,11 +76,25 @@ class MyWindow(urwid.Frame):
 		self.body.search_do(text, **flags)
 
 	def search_cancel(self):
-		self.footer = None
+		self.footer = self.help_line
 		self.focus_position = "body"
 
 	def show_error(self, msg: str) -> None:
-		self.footer = urwid.Text((ATTR_ERROR, msg))
+		self.errors.append(msg)
+		self.footer = urwid.Pile([urwid.Text((ATTR_ERROR, msg)) for msg in self.errors])
+
+	def reset_errors(self) -> None:
+		self.errors.clear()
+		self.footer = self.help_line
+
+	def keypress(self, size, key):
+		self.reset_errors()
+		return super().keypress(size, key)
+
+
+	@abc.abstractmethod
+	def get_selected_message(self) -> cantools.database.Message:
+		raise NotImplementedError()
 
 
 class MyListBox(urwid.ListBox):
@@ -128,7 +155,7 @@ class HexEdit(EditBaseClass):
 	fmt = '03X'
 
 	# override method of super class
-	def valid_char(ch: str) -> bool:
+	def valid_char(self, ch: str) -> bool:
 		return ch.lower() in '0123456789abcdef'
 
 	def check(self) -> bool:
@@ -143,16 +170,23 @@ class HexEdit(EditBaseClass):
 
 class UintEdit(EditBaseClass):
 
-	fmt = ''
+	def __init__(self, caption: str, *, default_value: 'float|None' = None, fmt: str = '') -> None:
+		super().__init__(caption=caption)
+		self.default_value = default_value
+		self.fmt = fmt
 
 	# override method of super class
 	def valid_char(self, ch: str) -> bool:
 		return ch.lower() in '0123456789'
 
 	def check(self) -> bool:
+		if self.default_value is not None:
+			return True
 		return bool(self.get_edit_text())
 
 	def value(self) -> int:
+		if self.default_value is not None and not self.get_edit_text():
+			return self.default_value
 		return int(self.get_edit_text())
 
 	def set_value(self, value: int) -> None:
@@ -379,28 +413,86 @@ class ChoiceBox(urwid.Overlay):
 
 class MessageList(MyWindow):
 
-	def __init__(self, db):
+	title = 'messages in {fn}'
+	_sort_key: 'Callable[[cantools.database.Message], object]'
+	_reverse: bool
+
+	def sort_by_name(self) -> object:
+		self._sort_key = lambda msg: msg.name
+		self._reverse = False
+		self.update()
+
+	def sort_by_id(self) -> object:
+		self._sort_key = lambda msg: msg.frame_id
+		self._reverse = False
+		self.update()
+
+	def sort_by_dbc(self) -> object:
+		self._sort_key = lambda msg: self.db.messages.index(msg)
+		self._reverse = False
+		self.update()
+
+	sortby_methods = {
+		'dbc': sort_by_dbc,
+		'id': sort_by_id,
+		'name': sort_by_name,
+	}
+
+	_sort_by_keys = tuple(sortby_methods.keys())
+	sortby = Config('message-list.sort-by', _sort_by_keys[0], allowed_values=_sort_by_keys)
+
+	_command_map = MyWindow._command_map.copy()
+	_command_map['0'] = 'sort_by_dbc'
+	_command_map['1'] = 'sort_by_id'
+	_command_map['2'] = 'sort_by_name'
+
+	def __init__(self, db: cantools.database.Database, ffn: str) -> None:
 		self.db = db
-		title = 'messages'
+		path, fn = os.path.split(ffn)
+		title = self.title.format(ffn=ffn, fn=fn)
 		shortcuts = [
 			('up/down', 'select message'),
 			('+', 'new message'),
 			('-', 'delete message'),
-			('enter', 'show signales'),
-			#TODO: edit message
-			('q', 'quit')
+			('enter', 'show signals'),
+			('i', 'edit message'),
+			('q', 'quit'),
+			('1', 'sort by id'),
+			('2', 'sort by name'),
+			('0', 'sort by dbc'),
 		]
 
 		super().__init__()
 		self.set_title(title)
 		self.set_help_line(shortcuts)
-		self.update()
+		self.sortby_methods[self.sortby](self)
 
 	def update(self):
-		content = [MyAttrMap(MessageLineWidget(msg)) for msg in self.db.messages]
+		content = [MyAttrMap(MessageLineWidget(msg)) for msg in sorted(self.db.messages, key=self._sort_key, reverse=self._reverse)]
 		content = MyListBox(content)
 		self.set_body(content)
 
+	def _sort_key_name(self, msg: cantools.database.Message) -> str:
+		return msg.name
+
+
+	def keypress(self, size, key: str) -> 'str|None':
+		if not super().keypress(size, key):
+			return None
+
+		cmd = self._command_map[key]
+		if cmd == 'sort_by_id':
+			self.sort_by_id()
+		elif cmd == 'sort_by_name':
+			self.sort_by_name()
+		elif cmd == 'sort_by_dbc':
+			self.sort_by_dbc()
+		else:
+			return key
+
+		return None
+
+
 	def get_selected_message(self):
 		return self.body.focus.base_widget.msg
 
@@ -432,25 +524,17 @@ class MessageLineWidget(LineWidget):
 
 class MessageEdit(MyWindow):
 
-	#TODO: check user input like in SignalEdit
-
 	title_edit = 'edit message {msg.frame_id:03x} {msg.name}'
 	title_new = 'create new message'
 
-	fmt_frame_id = Config('message-edit.fmt.frame_id', '{:03x}')
-	fmt_is_extended_frame = Config('message-edit.fmt.is_extended_frame', '{}')
-	fmt_length = Config('message-edit.fmt.length', '{}')
-	fmt_name = Config('message-edit.fmt.name', '{}')
-	fmt_comment = Config('message-edit.fmt.comment', '{}')
-
 	def __init__(self, db, msg=None, **kw):
 		self.db = db
 		self.msg = msg
-		self.edit_frame_id = urwid.Edit("Frame ID: ")
-		self.edit_is_extended_frame = urwid.Edit("Is extended: ")
-		self.edit_name = urwid.Edit("Message Name: ")
-		self.edit_comment = urwid.Edit("Comment: ")
-		self.edit_length = urwid.Edit("Data length (in bytes): ")
+		self.edit_frame_id = HexEdit("Frame ID: ")
+		self.edit_is_extended_frame = ChoiceEdit("Is extended: ", choices=dict(true=True, false=False))
+		self.edit_name = TextEdit("Message Name: ")
+		self.edit_comment = OptionalTextEdit("Comment: ")
+		self.edit_length = UintEdit("Data length (in bytes): ", default_value=0)
 
 		if msg:
 			title = self.title_edit.format(msg=msg)
@@ -458,9 +542,8 @@ class MessageEdit(MyWindow):
 				val = getattr(msg, key)
 				if val is None:
 					val = ''
-				val = getattr(self, 'fmt_%s' % key).format(val)
 				widget = getattr(self, 'edit_%s' % key)
-				widget.set_edit_text(val)
+				widget.set_value(val)
 		else:
 			title = self.title_new
 
@@ -478,17 +561,35 @@ class MessageEdit(MyWindow):
 
 	def get_input_data(self):
 		out = {}
-		out['frame_id'] = int(self.edit_frame_id.get_edit_text(), base=16)
-		out['is_extended_frame'] = self.edit_is_extended_frame.get_edit_text().strip().lower() != 'false'
-		out['name'] = self.edit_name.get_edit_text()
-		out['comment'] = self.edit_comment.get_edit_text()
-		out['length'] = 0 if not self.edit_length.get_edit_text() else int(self.edit_length.get_edit_text())
+		out['frame_id'] = self.edit_frame_id.value()
+		out['is_extended_frame'] = self.edit_is_extended_frame.value()
+		out['name'] = self.edit_name.value()
+		out['comment'] = self.edit_comment.value()
+		out['length'] = self.edit_length.value()
 		return out
 
+	def check(self, *, silent: bool = False) -> bool:
+		for w in (
+			self.edit_frame_id,
+			self.edit_is_extended_frame,
+			self.edit_name,
+			self.edit_comment,
+			self.edit_length,
+		):
+			if not w.check():
+				if not silent:
+					self.show_error('Invalid input for %s' % w.caption.rstrip().rstrip(':'))
+				return False
+
+		return True
+
 	def has_changed(self):
 		if not self.msg:
 			return True
 
+		if not self.check(silent=True):
+			return True
+
 		data = self.get_input_data()
 		for key in data:
 			new_val = data[key]
@@ -512,6 +613,9 @@ class MessageEdit(MyWindow):
 		return super().cmd_back()
 
 	def reply_save(self):
+		if not self.check():
+			return
+
 		if self.msg:
 			for attr, value in self.get_input_data().items():
 				setattr(self.msg, attr, value)
@@ -532,6 +636,48 @@ class MessageEdit(MyWindow):
 		return self
 
 
+	def get_selected_message(self) -> cantools.database.Message:
+		return self.msg
+
+
+class MessageLayout(MyWindow):
+
+	title = 'layout of message {msg.frame_id:03x} {msg.name}'
+
+	def __init__(self, msg, **kw):
+		self.msg = msg
+
+		import cantools.subparsers.dump.formatting
+		layout = cantools.subparsers.dump.formatting.layout_string(msg)
+		layout = '\n' + layout
+		body = [urwid.Text(ln, wrap='clip') for ln in layout.splitlines()]
+		body = MyListBox(body)
+
+		shortcuts = [
+			('up/down', 'scroll'),
+			('left', 'back'),
+			('q', 'quit')
+		]
+
+		super().__init__(body, **kw)
+		self.set_title(self.title.format(msg=msg))
+		self.set_help_line(shortcuts)
+
+
+	#def cmd_open(self):
+	#	return SignalList(self.msg, prev_window=self)
+
+	def cmd_edit(self):
+		return SignalList(self.msg, prev_window=self)
+
+	def cmd_cancel(self):
+		return self.cmd_back()
+
+
+	def get_selected_message(self) -> cantools.database.Message:
+		return self.msg
+
+
 class SignalList(MyWindow):
 
 	title = 'signals of message {msg.frame_id:03x} {msg.name}'
@@ -544,7 +690,7 @@ class SignalList(MyWindow):
 			('up/down', 'select signal'),
 			('+', 'new signal'),
 			('-', 'delete signal'),
-			#TODO: edit signal
+			('i', 'edit signal'),
 			('q', 'quit')
 		]
 		body = [MyAttrMap(SignalLineWidget(sig)) for sig in sorted(self.msg.signals, key=lambda sig: sig.start)]
@@ -576,6 +722,10 @@ class SignalList(MyWindow):
 		self.update()
 		self.app.notify_has_changed()
 
+
+	def get_selected_message(self) -> cantools.database.Message:
+		return self.msg
+
 class SignalLineWidget(LineWidget):
 
 	def __init__(self, signal):
@@ -809,6 +959,10 @@ class SignalEdit(MyWindow):
 		return self
 
 
+	def get_selected_message(self) -> cantools.database.Message:
+		return self.msg
+
+
 
 class ChoiceLineEdit(urwid.Columns):
 
@@ -878,11 +1032,14 @@ class ChoiceLineEdit(urwid.Columns):
 
 # ------- main classes -------
 
+class UserInputError(Exception):
+	pass
+
 class CanDatabaseEditor:
 
 	palette = [
 		(ATTR_TITLE, 'brown', 'default'),
-		(ATTR_HELP_LINE, 'dark gray', 'default'),
+		(ATTR_HELP_LINE, 'dark blue', 'default'),
 		(ATTR_ERROR, 'dark red', 'default'),
 	]
 
@@ -892,11 +1049,13 @@ class CanDatabaseEditor:
 	command_map['w'] = 'write'
 	command_map['q'] = 'quit'
 	command_map['e'] = 'edit'
+	command_map['i'] = 'edit'
 	command_map['esc'] = 'cancel'
 	command_map['+'] = 'new'
 	command_map['-'] = 'remove'
 	command_map['/'] = 'search'
 	command_map['?'] = 'search_reverse'
+	command_map['v'] = 'message_layout'
 
 	fallback_commands = {
 		'activate' : 'open',
@@ -904,14 +1063,20 @@ class CanDatabaseEditor:
 		'cursor left' : 'back',
 	}
 
+	# see what is supported by dump_file in
+	# https://github.com/cantools/cantools/blob/master/cantools/database/__init__.py
+	supported_database_formats = ('dbc', 'kcd', 'sym')
+
 	def __init__(self, fn=None):
 		self.init_focus_map()
-		self.filename = fn
 		self.has_changed = False
 		if fn and os.path.isfile(fn):
 			self.db = cantools.database.load_file(fn, prune_choices=False, sort_signals=None)
 		else:
+			if fn.split(os.path.extsep)[-1] not in self.supported_database_formats:
+				raise UserInputError('invalid database format, should be one of %s' % ', '.join(self.supported_database_formats))
 			self.db = cantools.database.Database()
+		self.filename = fn
 
 		MessageList.app = self
 		MessageEdit.app = self
@@ -919,6 +1084,9 @@ class CanDatabaseEditor:
 		SignalEdit.app = self
 		ChoiceBox.app = self
 
+		self.logging_handler = LoggingHandler(self)
+		logging.basicConfig(handlers=[self.logging_handler], format='%(message)s', level=logging.WARNING)
+
 	def init_focus_map(self):
 		for name, fg, bg in tuple(self.palette):
 			name_focused = name + '_focused'
@@ -929,11 +1097,11 @@ class CanDatabaseEditor:
 		self.palette.append(('default_focused', 'default,standout', 'default'))
 
 	def run(self):
-		widget = MessageList(self.db)
+		widget = MessageList(self.db, self.filename)
 		self.loop = urwid.MainLoop(widget, self.palette, handle_mouse=False, unhandled_input=self.unhandled_input)
 		self.loop.run()
 
-	def get_current_window(self):
+	def get_current_window(self) -> MyWindow:
 		return self.loop.widget
 
 	def open_window(self, window):
@@ -948,6 +1116,12 @@ class CanDatabaseEditor:
 		elif cmd == 'write':
 			self.write()
 			return
+		elif cmd == 'message_layout':
+			current_window = self.get_current_window()
+			if isinstance(current_window, MessageLayout):
+				return
+			self.open_window(MessageLayout(current_window.get_selected_message(), prev_window=current_window))
+			return
 
 		window = self.get_current_window()
 		func = getattr(window, 'cmd_%s' % cmd, None)
@@ -978,14 +1152,53 @@ class CanDatabaseEditor:
 		self.quit_without_write()
 
 	def write(self):
-		#TODO: check that db is valid by writing it to a tmp file and loading it
+		if not self.check():
+			return
+
 		cantools.database.dump_file(self.db, self.filename)
 		self.has_changed = False
 
+	def check(self) -> bool:
+		self.logging_handler.reset_errors()
+		with tempfile.TemporaryDirectory() as tmppath:
+			fn = os.path.join(tmppath, os.path.split(self.filename)[1])
+			cantools.database.dump_file(self.db, fn)
+			try:
+				cantools.database.load_file(fn, prune_choices=False, sort_signals=None)
+			except Exception as e:
+				self.show_error(str(e))
+				return False
+
+		if self.logging_handler.has_errors():
+			return False
+
+		return True
+
 	def quit_without_write(self):
 		raise urwid.ExitMainLoop()
 
 
+class LoggingHandler(logging.Handler):
+
+	def __init__(self, app: CanDatabaseEditor) -> None:
+		super().__init__()
+		self.app = app
+		self._has_errors = False
+
+	def emit(self, record: logging.LogRecord):
+		if not self.has_errors():
+			self.app.show_error(_('The following error(s) would occur when trying to load this database. Please fix them before saving it:'))
+		msg = self.format(record)
+		self.app.show_error(msg)
+		self._has_errors = True
+
+
+	def has_errors(self) -> bool:
+		return self._has_errors
+
+	def reset_errors(self) -> None:
+		self._has_errors = False
+
 
 class PrintVersion(argparse.Action):
 
@@ -1004,8 +1217,11 @@ def main(arglist: 'list[str]|None' = None) -> None:
 	p.add_argument('file', help='CAN bus database file, e.g. dbc or sym file')
 
 	args = p.parse_args(arglist)
-	dbeditor = CanDatabaseEditor(args.file)
-	dbeditor.run()
+	try:
+		dbeditor = CanDatabaseEditor(args.file)
+		dbeditor.run()
+	except UserInputError as e:
+		print(e)
 
 
 if __name__ == '__main__':
